/*
 *  Copyright (C) 2006
 *  Jiri Slaby <jirislaby@gmail.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 */

/*
 * We have a ringbuffer with three pointers. Two are bound with userspace:
 * wrptr by which we indicate (in the beginning of mmaped area) where ends our
 * request (userspace has to compute); rdptr, by which userspace tells us, it
 * just computed all material until this position. And finally one -- out of the
 * userspace -- special rdptr (call it krdptr), where we store, what kernelspace
 * catch to read yet. See the asciichart below to illustrate what the situation
 * might look like.
 *
 *    start of ring                  end of ring
 * 1. | krdptr rdptr                 wrptr     |
 * 2. | krdptr                 rdptr wrptr     |
 * 3. | krdptr                 rdptr wrptr     |
 *
 * In the case 1. we just request something to compute and woke the process to
 * do the work. Ok, it computes results and alter rdptr to be equal to value
 * of wrptr (2.). The rest is up to kernel space -- to read computed results and
 * increment krdptr until it reaches rdptr.
 * Obviously invariant is krdptr <= rdptr <= wrptr <= krdptr (modulo length of
 * buffer).
 */
#define DEBUG 1
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/kthread.h>
#include <linux/mm.h>
#include <linux/poll.h>
#include <linux/uaccess.h>
#include <linux/vmalloc.h>
#include <linux/wait.h>

#include "fpio.h"

#if FP_STAMP == 1 /* only x86 platforms */
#define FP_MEASURE	1
#endif

#define FP_NAME		"fp"

#define FP_RWBUF_CNT	((10 * PAGE_SIZE - sizeof(struct fp_stcom)) / \
		sizeof(struct fp_compute))
#define FP_RWBUF_SIZE	(FP_RWBUF_CNT * sizeof(struct fp_compute))

#define fp_incptr(ptr, fp) ((ptr) = ((ptr) + 1) % FP_RWBUF_CNT)

struct fp_soft {
	unsigned int major;
	struct cdev cdev;
	struct task_struct *tsk;
	wait_queue_head_t rdsleep, wrsleep;

	struct fp_stcom *rw_buffer;
	unsigned int krdptr;

	atomic_t open_cnt;
} fp;


/******************************************************************************
 * Thread work
 */
static int fp_thread(void *data)
{
	struct fp_soft *fp = data;
	unsigned int timeout;

	set_user_nice(current, -20);

	while (!kthread_should_stop()) {
		if (wait_event_interruptible(fp->wrsleep,
				fp->krdptr != fp->rw_buffer->rdptr ||
				kthread_should_stop()))
			break;

		timeout = 50;
		pr_debug("we were woken: %u/%u\n", fp->krdptr,
				fp->rw_buffer->rdptr);
		while (fp->krdptr != fp->rw_buffer->rdptr && --timeout) {
			pr_info(FP_NAME ": %u results in %llx"
#if FP_MEASURE == 1
				       	" (%lu ticks)"
#endif
					"\n",
					fp->krdptr, *(u64 *)&fp->rw_buffer->
					c[fp->krdptr].res
#if FP_MEASURE == 1
					, jiffies - 
					fp->rw_buffer->c[fp->krdptr].stamp
#endif
					);
			fp_incptr(fp->krdptr, fp);
		}
	}

	return 0;
}

/******************************************************************************
 * File ops
 */
static int fp_open(struct inode *inode, struct file *filp)
{
	struct fp_soft *fp = container_of(inode->i_cdev, struct fp_soft, cdev);
	unsigned long a, s = PAGE_ALIGN(FP_RWBUF_SIZE +
			sizeof(struct fp_stcom));

	if (atomic_inc_return(&fp->open_cnt) > 2) {
		atomic_dec(&fp->open_cnt);
		return -EBUSY;
	}

	filp->private_data = fp;

	if (atomic_read(&fp->open_cnt) > 1)
		return 0;

	fp->tsk = kthread_create(fp_thread, fp, "fp_thread");
	if (IS_ERR(fp->tsk)) {
		printk(KERN_ERR FP_NAME ": can't create computing thread\n");
		return -ENOMEM;
	}

	pr_debug("Allocating: %lu bufs of size %zu with overall size %lu, "
		"+%zu, rounded to %lu\n",
		FP_RWBUF_CNT, FP_RWBUF_SIZE, sizeof(struct fp_compute),
		sizeof(struct fp_stcom), s);
	fp->rw_buffer = vmalloc_32_user(s);
	if (fp->rw_buffer == NULL)
		goto err_thr;

	memset(fp->rw_buffer, 0xfe, s);
	fp->krdptr = fp->rw_buffer->rdptr = fp->rw_buffer->wrptr = 0;

	for (a = (unsigned long)fp->rw_buffer; s > 0; a += PAGE_SIZE,
			s -= PAGE_SIZE)
		SetPageReserved(vmalloc_to_page((void *)a));

	wake_up_process(fp->tsk);

	return 0;
err_thr:
	kthread_stop(fp->tsk);
	return -ENOMEM;
}

static int fp_release(struct inode *inode, struct file *filp)
{
	struct fp_soft *fp = filp->private_data;
	unsigned long a, s = PAGE_ALIGN(FP_RWBUF_SIZE +
			sizeof(struct fp_stcom));

	if (atomic_dec_return(&fp->open_cnt) > 0)
		return 0;

	kthread_stop(fp->tsk);

	for (a = (unsigned long)fp->rw_buffer; s > 0; a += PAGE_SIZE,
			s -= PAGE_SIZE)
		ClearPageReserved(vmalloc_to_page((void *)a));

	vfree(fp->rw_buffer);

	return 0;
}

static unsigned int fp_poll(struct file *filp, struct poll_table_struct *pt)
{
	struct fp_soft *fp = filp->private_data;
	unsigned int mask = 0;

	poll_wait(filp, &fp->rdsleep, pt);
	if (fp->rw_buffer->rdptr != fp->rw_buffer->wrptr)
		mask |= POLLIN | POLLRDNORM;

	return mask;
}

static int fp_mmap(struct file *filp, struct vm_area_struct *vma)
{
	struct fp_soft *fp = filp->private_data;

	return remap_vmalloc_range(vma, fp->rw_buffer, vma->vm_pgoff);
}

static ssize_t fp_read(struct file *filp, char __user *buf, size_t cnt,
		loff_t *off)
{
	struct fp_soft *fp = filp->private_data;
	struct fp_compute *fc;
	unsigned int a;

	for (a = 0; a < 450 /*&& fp->rw_buffer->wrptr != fp->krdptr*/; a++) {
		fc = &fp->rw_buffer->c[fp->rw_buffer->wrptr];
		fc->op = '*';
		fc->x = 12.0;
		fc->y = 12.1;
#if FP_MEASURE == 1
		fc->stamp = jiffies;
#endif
		fp_incptr(fp->rw_buffer->wrptr, fp);
		fc = &fp->rw_buffer->c[fp->rw_buffer->wrptr];
		fc->op = '+';
		fc->x = 12.0;
		fc->y = 12.1;
#if FP_MEASURE == 1
		fc->stamp = jiffies;
#endif
		fp_incptr(fp->rw_buffer->wrptr, fp);
	}
	wake_up_interruptible(&fp->rdsleep);

	*off = 0;

	return 0;
}

static long fp_uioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	struct fp_soft *fp = filp->private_data;

	if (_IOC_TYPE(cmd) != FPIOC_MAGIC || _IOC_NR(cmd) > FPIOC_MAXNR)
                return -ENOTTY;

	switch (cmd) {
	case FPIOC_BUFCNT:
		return put_user(FP_RWBUF_CNT, (__u32 *)arg);
	case FPIOC_DONE:
		wake_up_interruptible(&fp->wrsleep);
		break;
	default:
		return -ENOTTY;
	}

	return 0;
}

static struct file_operations fp_fops = {
	.owner = THIS_MODULE,
	.open = fp_open,
	.release = fp_release,
	.read = fp_read,
	.poll = fp_poll,
	.mmap = fp_mmap,
	.unlocked_ioctl = fp_uioctl,
};

/******************************************************************************
 * Init/deinit
 */
static int __init fp_init(void)
{
	int retval = -ENOMEM;
	dev_t dev;

	init_waitqueue_head(&fp.rdsleep);
	init_waitqueue_head(&fp.wrsleep);
	atomic_set(&fp.open_cnt, 0);

	retval = alloc_chrdev_region(&dev, 0, 1, FP_NAME);
	if (retval) {
		printk(KERN_ERR FP_NAME ": can't register character device\n");
		goto end;
	}
	fp.major = MAJOR(dev);

	cdev_init(&fp.cdev, &fp_fops);
	fp.cdev.owner = THIS_MODULE;

	retval = cdev_add(&fp.cdev, dev, 1);
	if (retval)
		goto err_unreg;

	pr_debug(FP_NAME ": module init OK\n");

	return 0;
err_unreg:
	unregister_chrdev_region(dev, 1);
end:
	return retval;
}

static void __exit fp_exit(void)
{
	cdev_del(&fp.cdev);

	unregister_chrdev_region(MKDEV(fp.major, 0), 1);

	pr_debug(FP_NAME ": module removed\n");

	return;
}

module_init(fp_init);
module_exit(fp_exit);

MODULE_AUTHOR("Jiri Slaby <jirislaby@gmail.com>");
MODULE_DESCRIPTION("Floating point compute driver");
MODULE_LICENSE("GPL");
