/*
 *  Copyright (C) 2005-2007 Jiri Slaby <jirislaby@gmail.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <poll.h>
#include <pthread.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/types.h>

/*#include <linux/futex.h>*/
#define FUTEX_WAIT		0
#define FUTEX_WAKE		1

/* why the hell isn't this defined? */
#include <sys/syscall.h>
#define futex(a, b, c, d, e, f) syscall(SYS_futex, a, b, c, d, e, f)

#define PHM_IN_LIBPHANTOM

/** @file libphantom.h */
#include "libphantom.h"
#include "phantom.h"

#define HISTORY_SIZE		200
#define MAX_TORQUE		30000
#define VELOCITY_AVG		30
#define GRADUAL_TIME		((1000000 / 16) * 3)

/*
 * THESE CONSTANTS ARE TAKEN FROM DOS DRIVER
 * and must be changed to match the phantom model
 */

/* Thermal Constants */
#define Ccu	385.0		/* specific heat of copper */
#define Cfe	434.0		/* specific heat of iron */
#define Rcu	2.34		/* electrical resistance of the armature */
#define Rcf	3.10		/* thermal resistance, armature - house */
#define Rfa	14.0		/* thermal resistance, housing - atm */
#define Tatm	25.0		/* atmospheric Temperature */
#define Tmax	100.0		/* shut down at this temperature(125 max!) */
#define ApT	(10.0/32768.0)	/* conversion factor Amps per Torque tick */

#define Mcu	0.0105		/* mass of the copper armature */
#define Mfe	0.120		/* mass of the iron motor housing */

#define print2(args...) fprintf(stderr, args)

#define to_pdev(dev)	(struct _phantom_device *)((char *)dev - \
		(size_t) &((struct _phantom_device *)0)->shared)

struct phantom_ang {
	short e[3];		/* phantom angles */
	short p[3];		/* stylus angles */
};

struct phantom_description {
	float k1, k2;
	float l1, l2;			/* lengths of arms */
	float cpr1;			/* motor counts per revolution */
	float cpr2;			/* gimbal encoders */
	float weight1, weight2;		/* counterweight */
	float atmos_temperature;
	float max_temperature;		/* shut down at this temperature(125 max!) */
	float max_torque;
	float mass_cu;			/* mass of the copper armature */
	float mass_fe;			/* mass of the iron motor housing */
	float scale;			/* scale factor for metric system */
};

struct _phantom_device {
	struct phantom_device shared;
	pthread_mutex_t mutex;

	struct timeval start;
	unsigned long frequency;
	unsigned int write_to;
	unsigned int ready;

	int fd;

	struct phantom_ang ang;
	float force[3];
	float tCu[3], tFe[3];			/* temperatures */
	float Tm[3];				/* torques */
	float avg_velocity[3];			/* average velocity sum */
	unsigned int axis;
	unsigned long max_diff;

	unsigned long hw_status;
	unsigned int is_stylus;
	__u32 control;

	struct phantom_pos actual;
	struct phantom_description desc;

	pthread_t comp_thread;
	int sleep;

	unsigned int stored_ptr;
	struct phantom_info info[2];
	struct phantom_pos stored[HISTORY_SIZE];

	unsigned char oh;
};

static struct _phantom_device ph_dev;

static void phantom_get(struct _phantom_device *dev, struct phantom_ang *ang)
{
	struct phm_regs regs = { .count = 7, .mask = 0xff };

	if (ioctl(dev->fd, PHN_GETREGS, &regs) < 0)
		perror("ioctl PHN_GETREGS failed");

#define ENCODER(axis) ((short)((0 - (int)regs.values[(axis)]) & 0xffff))
	ang->e[0] = ENCODER(0);
	ang->e[1] = ENCODER(1);
	ang->e[2] = ENCODER(2);

	ang->p[0] = ENCODER(3);
	ang->p[1] = ENCODER(4);
	ang->p[2] = ENCODER(5);
#undef ENCODER
	dev->hw_status = regs.values[6];
}

#ifdef USE_ROUNDING
static inline int myround(float x)
{
	return x < 0.0 ? x + 0.5 : x - 0.5;
}
#else
#define myround(x) (x)
#endif

static inline double sqr(double a)
{
	return a * a;
}

static void phantom_set_torque(struct _phantom_device *dev)
{
	struct phm_regs regs = { .count = 3, .mask = 0xff };
	unsigned int a;

	for (a = 0; a < 3; a++)
		regs.values[a] = (__s16)(myround(dev->Tm[a] / 16)) +
				PHN_ZERO_FORCE;

	if (ioctl(dev->fd, PHN_SETREGS, &regs) < 0)
		perror("ioctl set torques failed");

	if (dev->oh) {
		struct phm_reg reg = { .reg = PHN_CONTROL };

		dev->control ^= PHN_CTL_AMP;
		reg.value = dev->control;
		if (ioctl(dev->fd, PHN_SETREG, &reg) < 0)
			perror("ioctl change amp failed");
	}
}

static inline unsigned long timeval_diff(struct timeval *x, struct timeval *y)
{
	unsigned int nsec;

	/* Perform the carry for the later subtraction by updating y. */
	if (x->tv_usec < y->tv_usec) {
		nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
		y->tv_usec -= 1000000 * nsec;
		y->tv_sec += nsec;
	}
	if (x->tv_usec - y->tv_usec > 1000000) {
		nsec = (x->tv_usec - y->tv_usec) / 1000000;
		y->tv_usec += 1000000 * nsec;
		y->tv_sec -= nsec;
	}

	return (x->tv_sec - y->tv_sec) * 1000000 + x->tv_usec - y->tv_usec;
}

static void phm_error(struct _phantom_device *dev, unsigned int newstat)
{
	char buf[128];

	newstat &= PHM_ERROR;

	if (newstat && !(dev->shared.status & PHM_ERROR)) {
		dev->shared.status |= newstat;
		dev->shared.status &= ~PHM_RUNNING;
		strcpy(buf, "Stopping phantom due to");
		if (newstat & PHM_EOVERHEAT)
			strcat(buf, " OVERHEAT");
		if (newstat & PHM_EBIGFORCE)
			strcat(buf, " BIGFORCE");
		if (newstat & PHM_EBIGSPEED)
			strcat(buf, " BIGSPEED");
		strcat(buf, "!\n");
		print2("%s", buf);
	}
}

/*
 * Temperature check
 *
 * Integrate heat flow to get expected temperature of the device
 * Set OVERHEAT error bit if its too hot
 */
static void temperature_check(struct _phantom_device *dev)
{
	float Qcf, Qin, Qfa; /* heat flows */
	unsigned int axis = dev->axis;

	/* do one axis (motor) per step */
	dev->axis = (axis + 1) % 3;

	/* Qin - heat coming from the motor */
	Qin = ApT * dev->Tm[axis];
	Qin = Qin * Qin * Rcu;

	/* heat flow through copper part */
	Qcf = (dev->tCu[axis] - dev->tFe[axis]) / Rcf;
	/* heat flow through iron part */
	Qfa = (dev->tFe[axis] - dev->desc.atmos_temperature) / Rfa;

	dev->tCu[axis] += (Qin - Qcf) /
			(dev->desc.mass_cu * Ccu * dev->frequency / 3);

	dev->tFe[axis] += (Qcf - Qfa) /
			(dev->desc.mass_fe * Cfe * dev->frequency / 3);

	if (dev->tCu[axis] > Tmax)
		phm_error(dev, PHM_EOVERHEAT);
}

/* compute torques for hardware */
static void phantom_apply_force(struct _phantom_device *dev)
{
	if (dev->shared.status & PHM_RUNNING) {
		float fclg;

		/* using latest read values */
		float alpha = dev->desc.k1 * dev->ang.e[0];
		float beta = -dev->desc.k2 * dev->ang.e[1] - M_PI_2;
		float gamma = -dev->desc.k2 * dev->ang.e[2];

		/* precalculate all sin & cos, we'll need them all */
		float c_alpha = cos(alpha);
		float s_alpha = sin(alpha);
		float c_beta = cos(beta);
		float s_beta = sin(beta);
		float c_gamma = cos(gamma);
		float s_gamma = sin(gamma);

		float clg = dev->desc.l2 * c_gamma + dev->desc.l1 * c_beta;
		float slg = dev->desc.l2 * s_gamma + dev->desc.l1 * s_beta;

		/* handle startup to prevent big force - slow increase of
		   force */
		float gradual = dev->desc.scale;
		unsigned long d;

		d = timeval_diff(&dev->actual.stamp, &dev->start);
		if (d / 16 < GRADUAL_TIME)
			gradual *= (double)(d / 16) / (double)GRADUAL_TIME;

		/* convert forces to torques */
		fclg = s_alpha * dev->force[0] - c_alpha * dev->force[2];
		dev->Tm[0] = (-s_alpha * dev->force[2] -
				c_alpha * dev->force[0]) *
				dev->desc.k1 * clg * gradual;

		dev->Tm[1] = (s_beta * fclg + c_beta * dev->force[1]) *
			dev->desc.k2 * dev->desc.l1 * gradual;

		dev->Tm[2] = ((slg * fclg + clg * dev->force[1]) *
				dev->desc.k2 * gradual) - dev->Tm[1];

		/* compensate weight of the device -> ZERO FORCE */
		dev->Tm[1] -= dev->desc.weight1 * c_beta;
		dev->Tm[2] -= dev->desc.weight2 * c_gamma;

		//dev->force[0] *= 0.95;
		//dev->force[1] *= 0.95;
		//dev->force[2] *= 0.95;
		if (fabs(dev->Tm[0]) < dev->desc.max_torque &&
				fabs(dev->Tm[1]) < dev->desc.max_torque &&
				fabs(dev->Tm[2]) < dev->desc.max_torque)
			/* torques are within range -> apply */
			phantom_set_torque(dev);
		else
			phm_error(dev, PHM_EBIGFORCE);
	} else {
		/* zero torques */
		dev->Tm[0] = dev->Tm[1] = dev->Tm[2] = 0;
		phantom_set_torque(dev);
	}
}

/* do one step - read new data, compute position & velocity */
static void phantom_step(struct _phantom_device *dev)
{
	struct timeval now;
	float slg, clg;
	float v1, v2, v3;
	double timediff;
	unsigned long td;
	unsigned int i;

	/* read encoders */
	phantom_get(dev, &dev->ang);

	gettimeofday(&now, NULL);

	/* time differrence between the previous step */
	td = timeval_diff(&now, &dev->actual.stamp);
	memcpy(&dev->actual.stamp, &now, sizeof(now));

	/* look for largest differrence - statistical */
	if (dev->max_diff == 0)
		dev->max_diff++;	/* eliminate wrong maximum */
	else if (td > dev->max_diff)
		dev->max_diff = td;

	if (dev->hw_status & PHN_CTL_BUT)
		dev->shared.status &= ~PHM_BUTTON;
	else if (!(dev->shared.status & PHM_BUTTON)) {
		memcpy(&dev->shared.button, &now, sizeof(now));
		dev->shared.status |= PHM_BUTTON;
	}

	/* convert angles to position */
	dev->actual.alpha = dev->desc.k1 * -dev->ang.e[0];
	dev->actual.beta = dev->desc.k2 * -dev->ang.e[1];
	dev->actual.gamma = dev->desc.k2 * dev->ang.e[2];

	timediff = (double)td / 1e6; /* useconds */

	clg = dev->desc.l1 * sin(dev->actual.beta) +
			dev->desc.l2 * cos(dev->actual.gamma);
	slg = dev->desc.l1 * cos(dev->actual.beta) +
			dev->desc.l2 * sin(dev->actual.gamma);

	v1 = sin(dev->actual.alpha) * clg;
	dev->actual.speed[0] = (v1 - dev->actual.pos[0]) / timediff;
	dev->actual.pos[0] = v1;

	v2 = dev->desc.l2 - slg;
	dev->actual.speed[1] = (v2 - dev->actual.pos[1]) / timediff;
	dev->actual.pos[1] = v2;

	v3 = cos(dev->actual.alpha) * clg - dev->desc.l1;
	dev->actual.speed[2] = (v3 - dev->actual.pos[2]) / timediff;
	dev->actual.pos[2] = v3;

#define CONV (dev->desc.cpr2 / M_2_PI)

	/* pencil/gimball encoders */
	dev->actual.pos_enc[0] = -fmod(dev->ang.p[0] / CONV, M_2_PI);
	dev->actual.pos_enc[1] = fmod(dev->ang.p[1] / CONV, M_2_PI);
	dev->actual.pos_enc[2] = -fmod(dev->ang.p[2] / CONV, M_2_PI);

	/* copy to user buffer */
	if (dev->stored_ptr == ~0U) {
		unsigned int j;

		/* initialize buffer with first value */
		for (j = 0; j < HISTORY_SIZE; j++)
			memcpy(&dev->stored[j], &dev->actual,
				sizeof(dev->actual));

		for (j = 0; j < 3; j++)
			dev->avg_velocity[j] = VELOCITY_AVG *
				dev->actual.speed[j];
	}

	i = (dev->stored_ptr + 1) % HISTORY_SIZE;
	memcpy(&dev->stored[i], &dev->actual, sizeof(dev->actual));
	/* stored_ptr is always valid!! */
	dev->stored_ptr = i;

	/* velocity in m/s */
	v1 = sqr(dev->actual.speed[0]) +
		sqr(dev->actual.speed[1]) + sqr(dev->actual.speed[2]);
	if (v1 > 5*5 && dev->shared.iterations > 0)
		phm_error(dev, PHM_EBIGSPEED);

	/* in shared buffer is unfiltered speed */
	for (i = 0; i < 3; i++) {
		dev->avg_velocity[i] += dev->actual.speed[i];
		dev->avg_velocity[i] -= dev->stored[(dev->stored_ptr +
				(HISTORY_SIZE - VELOCITY_AVG)) %
					HISTORY_SIZE].speed[i];
		dev->actual.speed[i] = dev->avg_velocity[i] / VELOCITY_AVG;
	}

	//phantom_linear_regress(dev, 5, 5);
	//phantom_apply_force(dev);
	temperature_check(dev);
}

#if 0 /* unused for now, if somebody wants it, uncomment and add dev -> _dev */
/* set new description parameters
 * can be called only from RT environment with floats saved */
void phantom_set_description(struct _phantom_device *dev,
		struct phantom_description *des)
{
	/* check every parameter for reasonable value */

	if (des->l1 > 0.1 && des->l1 < 2.0)
		dev->desc.l2 = des->l1;

	if (des->l2 > 0.1 && des->l2 < 2.0)
		dev->desc.l2 = des->l2;

	if (des->k1 > 0.0 && des->k1 < 0.1)
		dev->desc.k1 = des->k1;

	if (des->k2 > 0.0 && des->k2 < 0.1)
		dev->desc.k2 = des->k2;

	if (des->cpr1 > 100 && des->cpr1 < 16384)
		dev->desc.cpr1 = des->cpr1;

	if (des->cpr2 > 100 && des->cpr2 < 16384)
		dev->desc.cpr2 = des->cpr2;

	if (des->weight1 > -3000 && des->weight1 < 3000)
		dev->desc.weight1 = des->weight1;

	if (des->weight2 > -3000 && des->weight2 < 3000)
		dev->desc.weight2 = des->weight2;

	if (des->atmos_temperature > -30 && des->atmos_temperature < 60)
		dev->desc.atmos_temperature = des->atmos_temperature;

	if (des->max_temperature > 30 && des->max_temperature < 120)
		dev->desc.max_temperature = des->max_temperature;

	if (des->max_torque > 1000 && des->max_torque < 25000)
		dev->desc.max_torque = des->max_torque;

	if (des->mass_cu > 0.01 && des->mass_cu < 0.03)
		dev->desc.mass_cu = des->mass_cu;

	if (des->mass_fe > 0.1 && des->mass_fe < 0.2)
		dev->desc.mass_fe = des->mass_fe;
}
#endif

static void phantom_stylus_detect(struct _phantom_device *dev)
{
	struct phantom_ang ang;
	unsigned int i;
	int j = 0;

	memset(&ang, 0, sizeof(ang));

	for (i = 0; i < 3; i++) {
		usleep(50000);
		phantom_get(dev, &ang);
		j |= ang.p[0] | ang.p[1] | ang.p[2];
	}

	dev->is_stylus = j != 0;
}

static void phantom_schedule(struct _phantom_device *dev)
{
	struct pollfd ufd = {
		.fd = dev->fd,
		.events = POLLIN
	};

	if (poll(&ufd, 1, -1) < 0)
		perror("poll failed");
	if (ufd.revents & POLLERR)
		print2("poll failed - POLLERR\n");
}

static void *phantom_compute(void *param)
{
	struct _phantom_device *dev = param;
	struct phantom_info *ph = ph;

	pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
	for (;;) {
		/* no write access to shared memory without this flag */
		if (dev->control & PHN_CTL_IRQ) {
			pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
			pthread_testcancel();
			phantom_schedule(dev);
			pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);

			/* apply computed values in last round, i.e. compute
			 * dev->Tm from dev->force and pass into the device */
			if (dev->shared.iterations) {
				memcpy(&dev->force, &ph->force,
						sizeof(dev->force));
				phantom_apply_force(dev);
			}

			/* convert new torques to pos and speed and store it in
			 * dev->actual */
			phantom_step(dev);

			pthread_mutex_lock(&dev->mutex);
			ph = &dev->info[dev->write_to];
			dev->shared.iterations++;
			memcpy(&ph->actual, &dev->actual, sizeof(dev->actual));
			dev->ready = dev->write_to;
			pthread_mutex_unlock(&dev->mutex);

			/* you have new data to compute, say phm_switch */
			futex(&dev->sleep, FUTEX_WAKE, 1, NULL, NULL,0);
		} else {
			pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
			pthread_testcancel();
			usleep(100000);
			pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
		}
	}

	return NULL;
}

/******************************************************************************
 * Init and deinit driver
 */

/* setup default parameters */
static void phantom_set_description_init(struct _phantom_device *dev)
{
	static const struct phantom_description des1_0 = {
		/* here are some number we don't know about how to use them

		ratio of capstain radius to cable disc radius

		// this should be aproximately
		K1 = R1 * (2 * M_PI / cpr1)
		K2 = R2 * (2 * M_PI / cpr1)

		Phantom 1.0
		R1 0.11507  => K1 0.0001765
		R2 0.13456  => K2 0.0002064

		Phantom 1.5
		R1 = 0.533 / 6.93
		R2 = 0.408 / 4.53

		R1 0.076912 => K1 0.0001180
		R2 0.090066 => K2 0.0001382

		Phantom 3.0
		R1 0.07587 => K1 0.00004692
		R2 0.08197 => K2 0.00005069

		*/
		/*
		!!! however there could be some connection with this value:

		ratio for digital torque values to motor output torque at the
		thimble

		Phantom 1.0 & 1.5
		#define Ttm 114000.0

		Phantom 3.0
		#define Ttm 23200.0
		*/
#if 0
		k1:0.0001752,	/* range for PI for rear motor */
		k2:0.0002045,	/* range for PI for two other motors */
#endif
		.k1 = 0.0001765,	/* range for PI for rear motor */
		.k2 = 0.0002064,	/* range for PI for two other motors */

		/* Lengths of arms and angle tranformations
		   these are the lengths of the PHANToM linkages in .001
		   inches
		   its about 0.1397m (5500 * 0.0254 / 1000) */
		/*
		Phantom 1.0  L1 = 5500.0,  L2 = 5500.0
		Phantom 1.5  L1 = 8250.0,  L2 = 8250.0
		Phantom 3.0  L1 = 18000.0, L2 = 18000.0
		*/
		.l1 = 0.13970,	/* in meters */
		.l2 = 0.13970,
		/* Motor counts per revolution
		newer - CPR1 4000
		Phantom 3.0 - CPR1 10160 */
		.cpr1 = 4096.0,

		/* Gimbal encoders */
		/* older CPR2 500 */
		.cpr2 = 2000.0,

#if 1
		/* Gimball & weight */
		.weight1 = 700.0,
		.weight2 = 1200.0,
		/* 700, 800 */
#endif

#if 0
		/* Gimball !weight */
		weight1:650.0,
		weight2:200.0,
#endif

#if 0
		/* Stylus !weight */
		weight1:-350.0,
		weight2:-1150.0,
		/* -500,-1060 */
#endif

#if 0
		/* Stylus & weight */
		weight1:-350.0,
		weight2:150.0,
#endif
		.atmos_temperature = Tatm,
		.max_temperature = Tmax,
		.max_torque = MAX_TORQUE,
		.mass_cu = Mcu,
		.mass_fe = Mfe,
		/* fixed 8bit decimal point
		* this almost looks like ~256 * 1024
		* however for now, this has been experimentaly calculated
		* containt gradual scale factor 256 */
		.scale = 256.0 * (6.97 * 100000 / 2.54)
					/* inches to meters to NEWTONS */
	};

	memcpy(&dev->desc, &des1_0, sizeof(dev->desc));
}

/**
 * wait until timer wakes us up
 *
 * This should be used in loop for waiting for new data. You probably might
 * want to call phm_switch after this returns.
 *
 * @param dev return value of phm_open
 */
void phm_wait(struct phantom_device *dev)
{
	struct _phantom_device *_dev = to_pdev(dev);

	futex(&_dev->sleep, FUTEX_WAIT, 0, NULL, NULL, 0);
}

/**
 * Open phantom device and create memory maps
 *
 * The complement, which should be called at the end of operation is phm_stop.
 *
 * @param devname phantom /dev device name
 * @return remember this pointer for passing it to further functions
 */
struct phantom_device *phm_open(const char *devname)
{
	struct _phantom_device *dev = &ph_dev;

	if (pthread_mutex_init(&dev->mutex, NULL)) {
		print2("phantom mutex init failed\n");
		goto err;
	}

	dev->fd = open(devname, O_RDWR);
	if (dev->fd < 0) {
		perror("phantom open failed");
		goto err_mutex;
	}

	dev->oh = ioctl(dev->fd, PHN_NOT_OH) < 0;
	if (dev->oh)
		print2("PHANTOM: using ohenhaptics (old) behaviour. maybe old "
			"kernel module?\n");

	return &dev->shared;
err_mutex:
	pthread_mutex_destroy(&dev->mutex);
err:
	return NULL;
}

/**
 * Interface to start force on phantom
 *
 * @param dev pointer returned by phm_open
 * @param unused no longer used, for backward compatibility
 * @return nonzero on failure, 0 if successful
 */
int phm_start(struct phantom_device *dev, unsigned int unused)
{
	struct phm_reg reg = { .reg = PHN_CONTROL };
	struct _phantom_device *_dev = to_pdev(dev);
	unsigned int a;

	if (unused)
		print2("PHANTOM: note, that rate (the second parameter of "
			"phm_start) is currently unused. Frequency is "
			"hardcoded to be about 1kHz. To avoid this warning "
			"pass zero rate.\n");

	/* initialize phantom */
	errno = 0;
	reg.value = _dev->control = PHN_CTL_IRQ | PHN_CTL_BUT;
	if (ioctl(_dev->fd, PHN_SETREG, &reg) < 0) {
		perror("ioctl setting irq failed");
		goto err;
	}

	phantom_set_description_init(_dev);
	for (a = 0; a < 3; a++)
		_dev->tCu[a] = _dev->tFe[a] = _dev->desc.atmos_temperature;
	
	phantom_stylus_detect(_dev);

	_dev->stored_ptr = ~0U;
	gettimeofday(&_dev->start, NULL);
	_dev->frequency = 1000;

	if (pthread_create(&_dev->comp_thread, NULL, phantom_compute, _dev)) {
		goto err_stop;
	}

	_dev->shared.status = PHM_ENABLED | PHM_RUNNING;

	return 0;
err_stop:
	reg.value = 0;
	ioctl(_dev->fd, PHN_SETREG, &reg);
err:
	return 1;
}

/**
 * stop phantom and release resources
 *
 * Call this, when you are finishing your work with phantom. It should be called
 * to be clean.
 *
 * It's safe to call this even if phm_start was not called, but phm_open was.
 *
 * @param dev pointer returned by phm_open
 */
void phm_stop(struct phantom_device *dev)
{
	struct phm_reg reg = { .reg = PHN_CONTROL };
	struct _phantom_device *_dev = to_pdev(dev);

	_dev->control &= ~PHN_CTL_IRQ;
	if (_dev->shared.status & PHM_ENABLED) {
		_dev->shared.status &= ~(PHM_ENABLED | PHM_RUNNING);
		pthread_cancel(_dev->comp_thread);
		pthread_join(_dev->comp_thread, NULL);
		ioctl(_dev->fd, PHN_SETREG, &reg);
	}
	close(_dev->fd);
	pthread_mutex_destroy(&_dev->mutex);
}

/**
 * reset phantom status
 *
 * This should be called before operating with phantom, since its position is
 * otherwise undefined and computing is bogus. Call this at least once after
 * first open.
 *
 * @param dev pointer returned by phm_open
 * @return nonzero on failure, 0 if successfully reset
 */
int phm_reset(struct phantom_device *dev)
{
	struct _phantom_device *_dev = to_pdev(dev);
	struct phm_reg reg = { .value = ~0x1f };

	if (ioctl(_dev->fd, PHN_SETREG, &reg))
		perror("ioctl reset PHN_SETREG failed");
	reg.value = 0x1f;
	if (ioctl(_dev->fd, PHN_SETREG, &reg))
		perror("ioctl reset PHN_SETREG failed");

	return 0;
}

/**
 * clamps the force vector to some maximum value
 *
 * @param force forces to alter
 * @param max_force what should be used as max force. If PHM_DEFAULT_MAX_FORCE,
 * use internal PHM_LIB_MAX_FORCE
 */
void phm_force_clamp_max(float force[3], float max_force)
{
	double f = sqr(force[0]) + sqr(force[1]) + sqr(force[2]);
	double m = (max_force == 0.0) ? sqr(PHM_LIB_MAX_FORCE) : sqr(max_force);

	if (f > m) {
		double t = sqrt(m / f);
		force[0] *= t;
		force[1] *= t;
		force[2] *= t;
	}
}

/**
 * get info struct where to write to and read from
 *
 * In most cases, this should be called *after* phm_wait returns.
 *
 * @param dev phantom_device which was returned by phm_open
 * @return actual info structure
 */
struct phantom_info *phm_switch(struct phantom_device *dev)
{
	struct _phantom_device *_dev = to_pdev(dev);
	unsigned int ready;

	pthread_mutex_lock(&_dev->mutex);
	ready = _dev->ready;
	_dev->write_to = !ready;
	pthread_mutex_unlock(&_dev->mutex);

	return &_dev->info[ready];
}
