#ifndef PHM_LIBPHANTOM_H
#define PHM_LIBPHANTOM_H

#include <math.h>
#include <string.h>

#include <sys/time.h>

#define PHM_LIB_MAX_FORCE	(12.0f)	/**< max force -- 12 Newtons */
#define PHM_DEFAULT_MAX_FORCE	(0.0f)	/**< use this to get default value
					     stored in the library */

#define PHM_EBIGSPEED	2		/**< speed overflew */
#define PHM_EBIGFORCE	4		/**< force overflew */
#define PHM_EOVERHEAT	8		/**< phantom is too hot */
/** some error occured */
#define PHM_ERROR	(PHM_EBIGSPEED|PHM_EBIGFORCE|PHM_EOVERHEAT)
/** is the button pressed? (don't use this directly, use phm_button_state
 * instead) */
#define PHM_BUTTON	16
#ifdef PHM_IN_LIBPHANTOM
#define PHM_ENABLED	32
#define PHM_RUNNING	64
#endif

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

/**
 * phantom_pos - actual pos contained in phantom_info
 */
struct phantom_pos {
	struct timeval stamp;		/**< timestamp */
	float pos[3];			/**< position of phantom arm */
	float pos_enc[3];		/**< stylus encoders - direct values */
	float alpha;			/**< angle on motors */
	float beta;
	float gamma;
	float speed[3];			/**< speed */
};

/**
 * phantom_info - position and force, returned by phm_switch
 *
 * There are 2 info structures in the libphantom internals, each one is
 * returned by phm_switch depending on time. There is no other way, how to get
 * known what is the actual pos and how to return froces back.
 */
struct phantom_info {
	struct phantom_pos actual;
	float force[3];
};

/**
 * phantom_device - returned by phm_open, contains global info
 *
 * It is passed to many functions, libphantom has ability to find out which
 * internal structure belongs to.
 */
struct phantom_device {
	unsigned long iterations;		/**< of a thread loop */
	unsigned int status;			/**< status of device */
	struct timeval button;			/**< button timestamp */
};

extern struct phantom_device *phm_open(const char *devname);
extern int phm_start(struct phantom_device *dev, unsigned int rate);
extern void phm_stop(struct phantom_device *dev);
extern int phm_reset(struct phantom_device *dev);
extern void phm_wait(struct phantom_device *dev);
extern void phm_force_clamp_max(float force[3], float max_force);
extern struct phantom_info *phm_switch(struct phantom_device *dev);

/**
 * phm_button_state - get button state
 *
 * @param dev device returned by phm_start
 * @return nonzero if the button is pressed
 */
static inline int phm_button_state(const struct phantom_device *dev)
{
	return dev->status & PHM_BUTTON;
}

/* spherical */

extern float *phm_to_spherical(float in[3], float out[3]);
extern float *phm_to_cartesian(float in[3], float out[3]);
extern float *phm_interpolate(float inB[3], float inE[3], float out[3],
		float alpha);

/* stylus */

/**
 * get 3x3 rotational matrix for stylus
 */
extern void phm_get_stylus_matrix(float *m,
		float a, float b, float x, float y, float z);

/**
 * transform vector by rotational matrix for stylus
 */
extern void phm_get_stylus_vector(float *vector_out,
		const float *m, const float *vector_in);

/* filter */

extern void phm_filter_force(float force_out[3], float force_in[3]);
extern void phm_filter_init(void);
extern void phm_filter_1D_recursive(float force_out[3], float alpha,
		float force_in[3], float force_old[3]);

static inline void phm_filter_2D_recursive(float force_out[3], float r, float k,
		float force_in[3], float force_old[3], float force_old1[3])
{
	unsigned int i;

	for (i = 0; i < 3; i++)
		force_out[i] = force_in[i] + 2 * r * cos(k) * force_old[i] -
			r * r * force_old1[i];

	memcpy(force_old1, force_old, sizeof(float) * 3);
	memcpy(force_old, force_out, sizeof(float) * 3);
}

static inline void phm_music_2D_recursive(float force_out[3],
		float k, float force_in[3], float force_old[3])
{
	unsigned int i;

	for (i = 0; i < 3; i++)
		force_out[i] = force_in[i] + k * force_in[i];

	memcpy(force_old, force_out, sizeof(float) * 3);
}

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
