// gmMatrix3.cc - 3x3 element matrix class
//
// libgm++: Graphics Math Library
// Ferdi Scheepers and Stephen F May
// 15 June 1994
// ----------------------------------
// modified: October 2002
// Pavel Kolcarek, kolcarek@fi.muni.cz
// 

#include "gm_mat3.h"
#include "gm_vec2.h"
#include "gm_vec3.h"

// private function: RCD
// - dot product of row i of matrix A and row j of matrix B

inline double RCD(const gmMatrix3 & A, const gmMatrix3 & B, int i, int j)
{
	return A[i][0] * B[0][j] + A[i][1] * B[1][j] + A[i][2] * B[2][j];
}

// private function: MINOR
// - MINOR of M[r][c]; r in {0,1,2}-{r0,r1}; c in {0,1,2}-{c0,c1}

inline double MINOR(const gmMatrix3 & M, int r0, int r1, int c0, int c1)
{
	return M[r0][c0] * M[r1][c1] - M[r1][c0] * M[r0][c1];
}

// CONSTRUCTORS

gmMatrix3::gmMatrix3()
{
	assign(0, 0, 0, 0, 0, 0, 0, 0, 0);
}

gmMatrix3::gmMatrix3(const gmMatrix3 & M)
{
	assign(M[0][0], M[0][1], M[0][2],
	       M[1][0], M[1][1], M[1][2],
	       M[2][0], M[2][1], M[2][2]);
}

gmMatrix3::gmMatrix3(double a00, double a01, double a02,
		     double a10, double a11, double a12,
		     double a20, double a21, double a22)
{
	assign(a00, a01, a02, a10, a11, a12, a20, a21, a22);
}

// ASSIGNMENT

gmMatrix3 & gmMatrix3::assign(double a00, double a01, double a02,
			      double a10, double a11, double a12,
			      double a20, double a21, double a22)
{
	m_[0][0] = a00;
	m_[0][1] = a01;
	m_[0][2] = a02;
	m_[1][0] = a10;
	m_[1][1] = a11;
	m_[1][2] = a12;
	m_[2][0] = a20;
	m_[2][1] = a21;
	m_[2][2] = a22;

	return *this;
}

gmMatrix3 & gmMatrix3::operator =(const gmMatrix3 & M)
{
	assign(M[0][0], M[0][1], M[0][2],
	       M[1][0], M[1][1], M[1][2],
	       M[2][0], M[2][1], M[2][2]);

	return *this;
}

// OPERATORS

gmMatrix3 & gmMatrix3::operator +=(const gmMatrix3 & M)
{
	m_[0][0] += M[0][0];
	m_[0][1] += M[0][1];
	m_[0][2] += M[0][2];
	m_[1][0] += M[1][0];
	m_[1][1] += M[1][1];
	m_[1][2] += M[1][2];
	m_[2][0] += M[2][0];
	m_[2][1] += M[2][1];
	m_[2][2] += M[2][2];

	return *this;
}

gmMatrix3 & gmMatrix3::operator -=(const gmMatrix3 & M)
{
	m_[0][0] -= M[0][0];
	m_[0][1] -= M[0][1];
	m_[0][2] -= M[0][2];
	m_[1][0] -= M[1][0];
	m_[1][1] -= M[1][1];
	m_[1][2] -= M[1][2];
	m_[2][0] -= M[2][0];
	m_[2][1] -= M[2][1];
	m_[2][2] -= M[2][2];

	return *this;
}

gmMatrix3 & gmMatrix3::operator *=(const gmMatrix3 & M)
{
	assign(RCD(*this, M, 0, 0), RCD(*this, M, 0, 1), RCD(*this, M, 0, 2),
	       RCD(*this, M, 1, 0), RCD(*this, M, 1, 1), RCD(*this, M, 1, 2),
	       RCD(*this, M, 2, 0), RCD(*this, M, 2, 1), RCD(*this, M, 2, 2));

	return *this;
}

gmMatrix3 & gmMatrix3::operator *=(double d)
{
	m_[0][0] *= d;
	m_[0][1] *= d;
	m_[0][2] *= d;
	m_[1][0] *= d;
	m_[1][1] *= d;
	m_[1][2] *= d;
	m_[2][0] *= d;
	m_[2][1] *= d;
	m_[2][2] *= d;

	return *this;
}

gmMatrix3 & gmMatrix3::operator /=(double d)
{
	double di = 1 / d;

	m_[0][0] *= di;
	m_[0][1] *= di;
	m_[0][2] *= di;
	m_[1][0] *= di;
	m_[1][1] *= di;
	m_[1][2] *= di;
	m_[2][0] *= di;
	m_[2][1] *= di;
	m_[2][2] *= di;

	return *this;
}

gmMatrix3 gmMatrix3::operator +(const gmMatrix3 & M) const
{
	return gmMatrix3(m_[0][0] + M[0][0], m_[0][1] + M[0][1],
			 m_[0][2] + M[0][2], m_[1][0] + M[1][0],
			 m_[1][1] + M[1][1], m_[1][2] + M[1][2],
			 m_[2][0] + M[2][0], m_[2][1] + M[2][1],
			 m_[2][2] + M[2][2]);
}

gmMatrix3 gmMatrix3::operator -(const gmMatrix3 & M) const
{
	return gmMatrix3(m_[0][0] - M[0][0], m_[0][1] - M[0][1],
			 m_[0][2] - M[0][2], m_[1][0] - M[1][0],
			 m_[1][1] - M[1][1], m_[1][2] - M[1][2],
			 m_[2][0] - M[2][0], m_[2][1] - M[2][1],
			 m_[2][2] - M[2][2]);
}

gmMatrix3 gmMatrix3::operator -() const
{
	return gmMatrix3(-m_[0][0], -m_[0][1], -m_[0][2],
			 -m_[1][0], -m_[1][1], -m_[1][2],
			 -m_[2][0], -m_[2][1], -m_[2][2]);
}

gmMatrix3 gmMatrix3::operator *(const gmMatrix3 & M) const
{
	return gmMatrix3(RCD(*this, M, 0, 0), RCD(*this, M, 0, 1),
			 RCD(*this, M, 0, 2), RCD(*this, M, 1, 0),
			 RCD(*this, M, 1, 1), RCD(*this, M, 1, 2),
			 RCD(*this, M, 2, 0), RCD(*this, M, 2, 1),
			 RCD(*this, M, 2, 2));
}

gmMatrix3 gmMatrix3::operator *(double d) const
{
	return gmMatrix3(m_[0][0] * d, m_[0][1] * d, m_[0][2] * d,
			 m_[1][0] * d, m_[1][1] * d, m_[1][2] * d,
			 m_[2][0] * d, m_[2][1] * d, m_[2][2] * d);
}

gmMatrix3 gmMatrix3::operator /(double d) const
{
	assert(!gmIsZero(d));

	double di = 1 / d;

	return gmMatrix3(m_[0][0] * di, m_[0][1] * di, m_[0][2] * di,
			 m_[1][0] * di, m_[1][1] * di, m_[1][2] * di,
			 m_[2][0] * di, m_[2][1] * di, m_[2][2] * di);
}

gmMatrix3 operator *(double d, const gmMatrix3 & M)
{
	return gmMatrix3(M[0][0] * d, M[0][1] * d, M[0][2] * d,
			 M[1][0] * d, M[1][1] * d, M[1][2] * d,
			 M[2][0] * d, M[2][1] * d, M[2][2] * d);
}

bool gmMatrix3::operator ==(const gmMatrix3 & M) const
{
	return (gmFuzEQ(m_[0][0], M[0][0]) && gmFuzEQ(m_[0][1], M[0][1]) &&
		gmFuzEQ(m_[0][2], M[0][2]) && gmFuzEQ(m_[1][0], M[1][0]) &&
		gmFuzEQ(m_[1][1], M[1][1]) && gmFuzEQ(m_[1][2], M[1][2]) &&
		gmFuzEQ(m_[2][0], M[2][0]) && gmFuzEQ(m_[2][1], M[2][1]) &&
		gmFuzEQ(m_[2][2], M[2][2]));
}

bool gmMatrix3::operator !=(const gmMatrix3 & M) const
{
	return !(*this == M);
}

gmVector3 gmMatrix3::operator *(const gmVector3 & v) const
{
	return gmVector3(m_[0][0] * v[0] + m_[0][1] * v[1] + m_[0][2] * v[2],
			 m_[1][0] * v[0] + m_[1][1] * v[1] + m_[1][2] * v[2],
			 m_[2][0] * v[0] + m_[2][1] * v[1] + m_[2][2] * v[2]);
}

gmVector3 operator *(const gmVector3 & v, const gmMatrix3 & M)
{
	return gmVector3(v[0] * M[0][0] + v[1] * M[1][0] + v[2] * M[2][0],
			 v[0] * M[0][1] + v[1] * M[1][1] + v[2] * M[2][1],
			 v[0] * M[0][2] + v[1] * M[1][2] + v[2] * M[2][2]);
}

// OUTPUT

inline ostream & operator <<(ostream & os, const gmMatrix3 & M)
{
	os << "| " << M[0][0] << " " << M[0][1] << " " << M[0][2] << " |" <<
	    endl;
	os << "| " << M[1][0] << " " << M[1][1] << " " << M[1][2] << " |" <<
	    endl;
	os << "| " << M[2][0] << " " << M[2][1] << " " << M[2][2] << " |" <<
	    endl;

	return os;
}

// OPERATIONS

gmMatrix3 gmMatrix3::transpose() const
{
	return gmMatrix3(m_[0][0], m_[1][0], m_[2][0],
			 m_[0][1], m_[1][1], m_[2][1],
			 m_[0][2], m_[1][2], m_[2][2]);
}

gmMatrix3 gmMatrix3::inverse() const
{
	assert(!isSingular());

	return adjoint() * gmInv(determinant());
}

gmMatrix3 gmMatrix3::adjoint() const
{
	gmMatrix3 A;

	A[0][0] =  MINOR(*this, 1, 2, 1, 2);
	A[0][1] = -MINOR(*this, 0, 2, 1, 2);
	A[0][2] =  MINOR(*this, 0, 1, 1, 2);
	A[1][0] = -MINOR(*this, 1, 2, 0, 2);
	A[1][1] =  MINOR(*this, 0, 2, 0, 2);
	A[1][2] = -MINOR(*this, 0, 1, 0, 2);
	A[2][0] =  MINOR(*this, 1, 2, 0, 1);
	A[2][1] = -MINOR(*this, 0, 2, 0, 1);
	A[2][2] =  MINOR(*this, 0, 1, 0, 1);

	return A;
}

double gmMatrix3::determinant() const
{
	return m_[0][0] * MINOR(*this, 1, 2, 1, 2) -
	       m_[0][1] * MINOR(*this, 1, 2, 0, 2) +
	       m_[0][2] * MINOR(*this, 1, 2, 0, 1);
}

bool gmMatrix3::isSingular() const
{
	return gmIsZero(determinant());
}

gmVector2 gmMatrix3::transform(const gmVector2 & v) const
{
	return gmVector2(v[0] * m_[0][0] + v[1] * m_[1][0] + m_[2][0],
			 v[0] * m_[0][1] + v[1] * m_[1][1] + m_[2][1]);
}

// TRANSFORMATION MATRICES

gmMatrix3 gmMatrix3::identity()
{
	return gmMatrix3(1, 0, 0, 0, 1, 0, 0, 0, 1);
}

gmMatrix3 gmMatrix3::scale(double x, double y)
{
	return gmMatrix3(x, 0, 0, 0, y, 0, 0, 0, 1);
}

gmMatrix3 gmMatrix3::translate(double x, double y)
{
	return gmMatrix3(1, 0, 0, 0, 1, 0, x, y, 1);
}

gmMatrix3 gmMatrix3::rotate(double angle)
{
	double sine = sin(gmRadians(angle));
	double cosine = cos(gmRadians(angle));

	return gmMatrix3( cosine, sine,   0,
			  -sine,  cosine, 0,
			  0,	  0,	  1);
}

// arbitrary axis rotation matrix -- along vector "v" by angle "angle"
gmMatrix3 gmMatrix3::arb_rotate(gmVector3 & v, double angle)
{
	// This function contributed by Erich Boleyn (erich@uruk.org) 

	double s, c;		// sine, cosine
	double one_c;		// 1 - cosine
	double xx, yy, zz, xy, yz, zx, xs, ys, zs;	// products of vector elements

	gmVector3 tvector(v);

	tvector.normalize();

	if (tvector.isZero())	// generate an identity matrix and return 
		return identity();

	//  Arbitrary axis rotation matrix.
	//
	//  This is composed of 5 matrices, Rz, Ry, T, Ry', Rz', multiplied
	//  like so:  Rz * Ry * T * Ry' * Rz'.  T is the final rotation
	//  (which is about the X-axis), and the two composite transforms
	//  Ry' * Rz' and Rz * Ry are (respectively) the rotations necessary
	//  from the arbitrary axis to the X-axis then back.  They are
	//  all elementary rotations.
	//
	//  Rz' is a rotation about the Z-axis, to bring the axis vector
	//  into the x-z plane.  Then Ry' is applied, rotating about the
	//  Y-axis to bring the axis vector parallel with the X-axis.  The
	//  rotation about the X-axis is then performed.  Ry and Rz are
	//  simply the respective inverse transforms to bring the arbitrary
	//  axis back to it's original orientation.  The first transforms
	//  Rz' and Ry' are considered inverses, since the data from the
	//  arbitrary axis gives you info on how to get to it, not how
	//  to get away from it, and an inverse must be applied.
	//
	//  The basic calculation used is to recognize that the arbitrary
	//  axis vector (x, y, z), since it is of unit length, actually
	//  represents the sines and cosines of the angles to rotate the
	//  X-axis to the same orientation, with theta being the angle about
	//  Z and phi the angle about Y (in the order described above)
	//  as follows:
	//
	//  cos ( theta ) = x / sqrt ( 1 - z^2 )
	//  sin ( theta ) = y / sqrt ( 1 - z^2 )
	//
	//  cos ( phi ) = sqrt ( 1 - z^2 )
	//  sin ( phi ) = z
	//
	//  Note that cos ( phi ) can further be inserted to the above
	//  formulas:
	//
	//  cos ( theta ) = x / cos ( phi )
	//  sin ( theta ) = y / sin ( phi )
	//
	//  ...etc.  Because of those relations and the standard trigonometric
	//  relations, it is pssible to reduce the transforms down to what
	//  is used below.  It may be that any primary axis chosen will give the
	//  same results (modulo a sign convention) using thie method.
	//
	//  Particularly nice is to notice that all divisions that might
	//  have caused trouble when parallel to certain planes or
	//  axis go away with care paid to reducing the expressions.
	//  After checking, it does perform correctly under all cases, since
	//  in all the cases of division where the denominator would have
	//  been zero, the numerator would have been zero as well, giving
	//  the expected result.

	s = sin(gmRadians(angle));
	c = cos(gmRadians(angle));

	xx = tvector[0] * tvector[0];
	yy = tvector[1] * tvector[1];
	zz = tvector[2] * tvector[2];
	xy = tvector[0] * tvector[1];
	yz = tvector[1] * tvector[2];
	zx = tvector[2] * tvector[0];
	xs = tvector[0] * s;
	ys = tvector[1] * s;
	zs = tvector[2] * s;

	one_c = 1.0 - c;

	return gmMatrix3((one_c * xx) + c, (one_c * xy) - zs, (one_c * zx) + ys,
			 (one_c * xy) + zs, (one_c * yy) + c, (one_c * yz) - xs,
			 (one_c * zx) - ys, (one_c * yz) + xs,
			 (one_c * zz) + c);
}
