//-----------------------------------------------------------//
//							     //
//       ph_lib_data_structs.h	       ver. 3.0		     //
//							     //
//       Data structures used for computation of force.      //
//       Data itself can be obtained from other		     //
//       libraries like mr_lib, etc.			     //
//							     //
//       Pavel Kolcarek, kolcarek@fi.muni.cz		     //
//							     //
//-----------------------------------------------------------//

#ifndef __PH_LIB_DATA_STRUCTS_H__
#define __PH_LIB_DATA_STRUCTS_H__

#include <vector>

#include <libphantom.h>

#include "gm.h"
#include "ph_lib_def.h"
#include "mr_thread.h"

  ////////////////////////////////////////////////////////////
 //
//   class CPhTriangle
//   --------------------------------------------------------
//   The basic primitive, similar to MrTriangle, but independent
//   on mr_lib -- so that we can use ph_lib anywhere we
//   want :-)
//

class CPhTriangle {	// struct should be enough; if methods are
			// needed, it should be redefined to class
public:
	/* TODO */// -- nebylo by lepsi to mit jako pointery ????
	gmVector3 vertex[3];
	gmVector3 normal;	// triangle normal
	gmVector3 vertex_normal[3];

	gmVector3 active_point;	// real position of the PHANToM arm within
				// triangle
	bool active;		// is this triangle used for computation of
				// a force?
	bool has_real_normal;	// has triangle defined real normal or is the
				// normal computed only as a linear combination
				// of vertex normals?

	CPhTriangle():active(false), has_real_normal(false) {};
	CPhTriangle(gmVector3 &, gmVector3 &, gmVector3 &,
		     gmVector3 &, gmVector3 &, gmVector3 &, gmVector3 &);
};

  ////////////////////////////////////////////////////////////
 //
//   class CPhTriangleSet
//   --------------------------------------------------------
//   For haptic rendering we are using two sets of triangles.
//   The "active" set is used for computation of final force
//   and into the "not active" one new triangles are
//   inserted. This class implements these sets.
//

class CPhTriangleSet {
public:
	int count;		// current number of triangles in this set

	volatile bool is_ready;	// is this set filled in with correct data
	// and ready to be used?

	// vector of triangles used for computation of force
	vector < CPhTriangle * >triangles;

	CPhTriangleSet();
	~CPhTriangleSet();
};

  ////////////////////////////////////////////////////////////
 //
//   class CPhantomRec
//   --------------------------------------------------------
//   The class for accessing the current position, speed and
//   force of phantom. From this structure is computed
//   a current force.
//

class CPhantomRec {
private:
	gmVector3 ph_dev_position;	// vector of last read real finger
					// position from PHANToM device

	gmVector3 ph_position;	// vector of last known finger position
				// (usually a virtual one, computed by
				//  methods of phantom thread)

	gmVector3 ph_speed;	// vector of last known speed read from PHANToM
	gmVector3 ph_force;	// current force vector

	CPhTriangleSet *triangle_set[2];	// we use two sets of triangles
	int active_set;		// 0 or 1

	// features and attributes of the PHANToM
	double ph_force_radius;
	double ph_min_radius;
	double ph_max_force;
	double ph_force_delta;
	double ph_force_delta_virtual;
	bool ph_attrition;
	bool ph_button_state;
	gmVector3 ph_stylus_direction;
	gmMatrix4 ph_stylus_rotate;
	struct phantom_device *ph_global;

public:
	CPhantomRec();		// constructor & destructor
	~CPhantomRec();

	// this function get all necessary information we need --
	// position, speed and force. other attributes are ignored.
	// if phantom is not present, NULL is returned
	void update_phantom_info(struct phantom_info *phantom_rec);
	// this function sets to the internal phantom structure
	// __only__ actual force, we don't need anything more... (at this
	// moment?) if phantom is not present, nothing will happen.
	void set_phantom_info(struct phantom_info *phantom_rec);

	// attributes accessing stuff
	inline gmVector3 get_device_position()	{ return ph_dev_position; }
	inline gmVector3 get_position()		{ return ph_position; }
	inline gmVector3 get_speed()		{ return ph_speed; }
	inline double get_force_radius()	{ return ph_force_radius; }
	inline double get_min_radius()		{ return ph_min_radius; }
	inline double get_max_force()		{ return ph_max_force; }
	inline double get_force_delta()		{ return ph_force_delta; }
	inline double get_force_delta_virtual()	{return ph_force_delta_virtual;}
	inline bool get_button_state()		{ return ph_button_state; }
	inline gmVector3 get_stylus_direction()	{ return ph_stylus_direction; }
	inline gmMatrix4 get_stylus_rotate()	{ return ph_stylus_rotate; }
	inline void set_global_ptr(struct phantom_device *dev) {
		ph_global = dev;
	}

	inline bool is_attrition()		{ return ph_attrition; }

	inline void set_force(gmVector3 f)	{ ph_force = f; }
	inline void set_position(gmVector3 p)	{ ph_position = p; }
	inline void set_force_radius(double r)	{ ph_force_radius = r; }
	inline void set_min_radius(double r)	{ ph_min_radius = r; }
	inline void set_max_force(double f)	{ ph_max_force = f; }
	inline void set_force_delta(double d)	{ ph_force_delta = d; }
	inline void set_force_delta_virtual(double d) {
		ph_force_delta_virtual = d;
	}
	inline void set_attrition(bool a)	{ ph_attrition = a; }

	// methods for manipulation with sets of triangles
	CPhTriangleSet *get_active_set();
	CPhTriangleSet *get_non_active_set();
	bool switch_sets();	// switch active and non-active set of triangles
	// if non-active set is ready, switching
	// passes and true is returned
};

#endif
