//-----------------------------------------------------------//
//							     //
//	ph_lib_thread.cpp		   ver. 3.0	     //
//							     //
//	Implementation of CPhantomThread class.		     //
//	Its more precise description is in the header file.  //
//							     //
//	Pavel Kolcarek, kolcarek@fi.muni.cz		     //
//							     //
//-----------------------------------------------------------//

//#include <stdlib.h>
#include <iostream>
#include <unistd.h>
#include "ph_lib_lowlevel.h"
#include "ph_lib_thread.h"

using namespace std;

  ////////////////////////////////////////////////////////////
 //
//   class CPhantomThread

CPhantomThread::CPhantomThread(int ph_rate = DEFAULT_RATE)
{
	// phantom thread identification
	thread_id = 0;

	// set attributes to its init. values
	loop_is_running = false;
	wanna_quit_loop = false;
	max_force = DEFAULT_MAX_FORCE;
	phantom_rate = ph_rate;

#ifdef PHANTOM_PRESENT
	ph_open(phantom_rate, &ph_global);
	phantom_rec.set_global_ptr(ph_global);
#else
	cout << "|--> (phantom device support is OFF, PHANToM is not "
		"initialized)" << endl;
	cout << "|--> (application is running in TESTING MODE!)" << endl;
#endif
	// initialize mutex
	pthread_mutex_init(&thread_lock, NULL);

//	ptr_model_thread = NULL;
}

CPhantomThread::~CPhantomThread()
{
#ifdef PHANTOM_PRESENT
	ph_close();
#endif
	// and destroy lock
	pthread_mutex_destroy(&thread_lock);
}

// main thread loop function
void CPhantomThread::run_thread()
{
#ifdef PHANTOM_PRESENT
	// actual data from the phantom
	struct phantom_info *ph_rec;
#ifdef DEBUG
	static int passes = 1;
#endif				// DEBUG

#endif				// PASSES

	// a force exerted to the user
	gmVector3 tmp_force;

	do {
#ifdef PHANTOM_PRESENT
		st_wait(&ph_rec);	// wait for new data and get it
#endif

		lock_me();
#ifdef PHANTOM_PRESENT
		// store obtained data from st_wait() in phantom_rec
		phantom_rec.update_phantom_info(ph_rec);

		// compute the correct force for the active set of triangles
		tmp_force = compute_force(phantom_rec.get_active_set(),
				max_force);

		// apply all necessary filters (e.g. friction, max force, etc.)
		apply_filters(tmp_force);
		apply_additional_filters(tmp_force);

		// store the tmp_force in phantom_rec
		phantom_rec.set_force(tmp_force);
		// and send the stored force in phantom_rec to the phantom 
		// device's ph_rec
		phantom_rec.set_phantom_info(ph_rec);

		// and update the position and the virtual position
		old_device_position = phantom_rec.get_device_position();
		old_virtual_position = virtual_position;
		phantom_rec.set_position(virtual_position);

#ifdef DEBUG
		if (passes >= 1000) {
			cout << "*tisic* pruchodu PHANToM threadem" << endl;
			passes = 1;
		}
		passes++;
#endif

		// and finally switch active and non-active sets if
		// new data is ready
		phantom_rec.switch_sets();

#endif				// PHANTOM_PRESENT
		unlock_me();

	} while (!wanna_quit_loop);

	// we want to quit this thread, so let's prepare for it!
	// set the loop_is_running flag...
	lock_me();
	loop_is_running = false;
	unlock_me();

	// zeroize force...
	tmp_force.assign(0.0, 0.0, 0.0);
#ifdef PHANTOM_PRESENT
	phantom_rec.set_force(tmp_force);
	phantom_rec.set_phantom_info(ph_rec);
	unlock_me();
	st_wait(&ph_rec);
#endif

	// and quit... (this function ;-)
}

void CPhantomThread::start_phantom_thread()
{
	lock_me();
	wanna_quit_loop = false;
	unlock_me();

	if (!is_running()) {
		loop_is_running = true;

		// create new thread
		start_loop();
	}
#ifdef DEBUG
#ifdef PHANTOM_PRESENT
	cout << "|--> phantom thread started... force is ON :-)" << endl;
#else
	cout <<
	    "|--> phantom thread started... however, PHANToM is *NOT* present :-( force is OFF"
	    << endl;
#endif
#endif
}

void CPhantomThread::stop_phantom_thread()
{
	lock_me();
	wanna_quit_loop = true;
	unlock_me();

	// wait until current iteration finishes
	while (is_running()) ;

#ifdef DEBUG
	cout << "|--> phantom thread stopped... force is OFF :-)" << endl;
#endif
}

void CPhantomThread::filter_add_friction(gmVector3 & tmp_force)
{
	double dist_coef;
	gmVector3 vector;
	gmVector3 normal;
	gmVector3 speed;
	gmVector3 zero(0, 0, 0);

	if (max_force > 0 && phantom_rec.is_attrition()) {
		vector = old_virtual_position - virtual_position;
		normal = tmp_force.normalize();

		dist_coef = plane_distance(vector, zero, normal);
		vector = normal * dist_coef + vector;
		if (!gmIsZero(vector.length()))
			vector.normalize();

		speed = phantom_rec.get_speed();
		dist_coef = speed.length() * PH_FORCE_DELTA_ATTRITION;

		if (dist_coef > PH_FORCE_MAX_ATTRITION)
			dist_coef = PH_FORCE_MAX_ATTRITION;

		vector *= dist_coef;
		tmp_force += vector;

		if (dist_coef > max_force)
			max_force = dist_coef;
	}
}

void CPhantomThread::filter_crop_force(gmVector3 & tmp_force)
{
	tmp_force.normalize();

	if (max_force > phantom_rec.get_max_force())
		max_force = phantom_rec.get_max_force();

	tmp_force *= max_force;
}

void CPhantomThread::apply_filters(gmVector3 & tmp_force)
{
	filter_add_friction(tmp_force);
	// if force vector doesn't have zero lenght, crop it to fit max. force
	if (!gmIsZero(tmp_force.length())) {
		filter_crop_force(tmp_force);
	}
}

#ifdef THIS_IS_DEFUNCT
// -----------------------------
void CPhantomThread::apply_additional_filters(gmVector3 & tmp_force)
{
	for (unsigned int i = 0; i < additional_filter.size(); i++) {
		additional_filter[i] (tmp_force);
	}
//	if (!gmIsZero(tmp_force.length())) filter_crop_force(tmp_force);
}

void CPhantomThread::add_filter(unsigned int pos,
				void (*filterFunc) (gmVector3 &))
{
	std::vector < void (*) (gmVector3 &) >::iterator itr;

	if (additional_filter.size() > pos + 1)
		additional_filter.insert(additional_filter.end(), filterFunc);
	else {
		itr = additional_filter.begin();
		for (unsigned int i = 0; i < pos; i++)
			itr++;
		additional_filter.insert(itr, filterFunc);
	}
}

void CPhantomThread::add_filter_to_begin(void (*filterFunc) (gmVector3 &))
{
	additional_filter.insert(additional_filter.begin(), filterFunc);
}

void CPhantomThread::add_filter_to_end(void (*filterFunc) (gmVector3 &))
{
	additional_filter.insert(additional_filter.end(), filterFunc);
}

void CPhantomThread::remove_filter(unsigned int pos)
{
	std::vector < void (*) (gmVector3 &) >::iterator itr;

	if (additional_filter.size() < pos + 1)
		return;
	itr = additional_filter.begin();
	for (unsigned int i = 0; i < pos; i++)
		itr++;
	additional_filter.erase(itr);
}
#endif

// ------------------------------

void CPhantomThread::enable_force()
{
	/* TODO */
}

void CPhantomThread::disable_force()
{
	/* TODO */
}
